# typing.py
#
# Copyright 2025 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import IntEnum, auto
from typing import TypedDict

from gi.repository import Gio


class SettingsSource(TypedDict):
    """Dictionary for storing important `Gio.Settings`-related data."""

    settings: Gio.Settings | None
    schema: Gio.SettingsSchema | None


class LogicalDirectionType(IntEnum):
    """Logical direction types."""

    START = auto()
    END = auto()



#!@PYTHON@

# refine.in
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gettext
import locale
import os
import signal
import subprocess
import sys
from pathlib import Path

VERSION = "@VERSION@"
pkgdatadir = "@pkgdatadir@"
localedir = "@localedir@"

sys.path.insert(1, pkgdatadir)
signal.signal(signal.SIGINT, signal.SIG_DFL)
locale.bindtextdomain("refine", localedir)
locale.textdomain("refine")
gettext.install("refine", localedir)

# ruff: noqa: E402
from refine.logger import logger

if __name__ == "__main__":
    import gi

    gi.require_version("Xdp", "1.0")

    from gi.repository import GLib, Gio, Xdp

    refine_resource = Gio.Resource.load(str(Path(pkgdatadir, "refine.gresource")))
    Gio.resources_register(refine_resource)
    data_resource = Gio.Resource.load(str(Path(pkgdatadir, "data.gresource")))
    Gio.resources_register(data_resource)

    from refine import main

    if Xdp.Portal.running_under_flatpak():
        logger.debug("Running under Flatpak")
        output = subprocess.run(
            ("flatpak-spawn", "--host", "sh", "-c", 'echo "$XDG_DATA_DIRS"'),
            capture_output=True,
            text=True,
        )
        host_xdg_data_dirs = {
            Path(paths) for paths in output.stdout.strip().split(os.pathsep)
        }
        home = Path(GLib.get_home_dir()).name

        xdg_data_dirs = os.environ["XDG_DATA_DIRS"].split(os.pathsep)
        try:
            redundant_path = str(Path(os.sep, "usr", "share"))
            xdg_data_dirs.remove(redundant_path)
            os.environ["XDG_DATA_DIRS"] = os.pathsep.join(xdg_data_dirs)
            logger.debug(f"Removed “{redundant_path}” from XDG_DATA_DIRS")
        except ValueError:
            pass

        for directory in host_xdg_data_dirs:
            host_path = Path(os.sep, "run", "host", directory.relative_to(os.sep))
            host_schemas_path = host_path / "glib-2.0" / "schemas"

            if not host_schemas_path.exists():
                logger.debug(f"Ignoring “{str(directory).replace(home, '[REDACTED]')}”")
                continue

            xdg_data_dirs = os.environ["XDG_DATA_DIRS"].split(os.pathsep)
            xdg_data_dirs.insert(0, str(host_path))

            os.environ["XDG_DATA_DIRS"] = os.pathsep.join(xdg_data_dirs)
            logger.debug(f"Added “{host_path}” into XDG_DATA_DIRS")

    sys.exit(main.main())



# flags_switch_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import contextlib
from typing import Any, cast

from gi.repository import Gtk

from refine.widgets.base import InvalidSettingError
from refine.widgets.string_row import StringRow


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/flags-switch-row.ui")
class FlagsSwitchRow(StringRow):
    """A `ActionRow` to enable or disable a feature from a string list."""

    __gtype_name__ = "FlagsSwitchRow"

    def get_string(self) -> list[str]:
        """Retrieve list from key."""
        return self.settings.get_strv(self.key)

    def active_set(self, *_args: Any) -> None:
        """Set whether @self is in its “on” or “off” position."""
        flags = self.get_string()

        value = cast("str", self.value)
        if self.active:
            flags.append(value)
        else:
            with contextlib.suppress(ValueError):
                flags.remove(value)

        flags = list(set(flags))

        self.settings.set_strv(self.key, flags)
        self.value_set(flags)
        self.do_update()

    def do_update(self) -> None:
        super().do_update()
        schema_key = self.schema.get_key(self.key)

        try:
            key_range = schema_key.get_range()[1]

            if self.value not in key_range:
                raise InvalidSettingError
        except (IndexError, TypeError) as error:
            raise InvalidSettingError from error



# value_combo_row.py
#
# Copyright 2025 Mitchell Winkelman (Heliguy)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


from gi.repository import GObject, Gtk

from refine.widgets.base import InvalidSettingError
from refine.widgets.combo_row import ComboRow

MINIMUM_CHILDREN = 2


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/value-combo-row.ui")
class ValueComboRow(ComboRow):
    """An `Adw.ComboRow` implementing the properties needed to work with enumerated properties in `Gio.Settings`."""

    __gtype_name__ = "ValueComboRow"

    revert_button = Gtk.Template.Child()
    revealer = Gtk.Template.Child()

    title_case = GObject.Property(type=bool, default=False)
    omit_hyphens = GObject.Property(type=bool, default=False)
    omit_words = GObject.Property(type=str)

    def do_setup(self) -> None:
        enum_tuple: tuple[str, list[str]] = (
            self.schema.get_key(self.key).get_range().unpack()
        )
        if len(enum_tuple) < MINIMUM_CHILDREN or enum_tuple[0] != "enum":
            raise InvalidSettingError

        self.original_list = enum_tuple[1]

        self.transformed_list = self.original_list

        if self.omit_words:
            for word in self.omit_words.split():
                self.transformed_list = [
                    string.replace(word.lower(), "") for string in self.transformed_list
                ]

        if self.omit_hyphens:
            self.transformed_list = [
                string.replace("-", " ") for string in self.transformed_list
            ]

        if self.title_case:
            self.transformed_list = [string.title() for string in self.transformed_list]

        self.transformed_list = [string.strip() for string in self.transformed_list]

        combined_lists = sorted(
            zip(self.transformed_list, self.original_list, strict=False)
        )
        self.transformed_list = [string[0] for string in combined_lists]
        self.original_list = [string[1] for string in combined_lists]

        model = Gtk.StringList.new(self.transformed_list)
        string: str = self.settings.get_string(self.key)
        self.set_model(model)
        self.set_selected(self.original_list.index(string))

        self.set_is_default(self.key)
        self.connect("notify::selected", self.selected_callback)
        super().do_setup()



# logger.py
#
# Copyright (C) 2025 The Upscaler Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging
import os
import sys

from refine import PROFILE


class PackageNameFilter(logging.Filter):
    """The path name of the package."""

    def filter(self, record: logging.LogRecord) -> bool | logging.LogRecord:
        """Get the path name of the package."""
        package_absolute_path = sys.path[1] + os.sep
        record.packagename = record.pathname.replace(package_absolute_path, "")
        return True


logger = logging.getLogger()
format = "%(packagename)s:%(lineno)d:0:\n%(levelname)8s: %(funcName)s(): %(message)s"

if PROFILE == "development":
    logging.basicConfig(
        format=format,
        level=logging.DEBUG,
    )
else:
    logging.basicConfig(
        format=format,
    )
logger.handlers[0].addFilter(PackageNameFilter())



# base.py
#
# Copyright 2025 Hari Rana <theevilskeleton@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import abstractmethod
from typing import Any

from gi.repository import GLib, GObject, Gio, Gtk

from refine.logger import logger


class InvalidSettingError(TypeError): ...


class BaseInterface(GObject.Object):
    __gtype_name__ = "BaseInterface"

    key: str
    schema_id: str
    schema: Gio.SettingsSchema
    settings: Gio.Settings
    is_valid_setting: bool

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        GLib.idle_add(self._post_init)

    @staticmethod
    def set_up_separator_revealer(revealer: Gtk.Revealer, widget: Gtk.Widget) -> None:
        """Initialize and set up the `Gtk.Revealer`."""

        def state_flags_changed(
            widget: Gtk.Widget, _flags: Gtk.StateFlags, revealer: Gtk.Revealer
        ) -> None:
            flags = widget.get_state_flags()

            is_prelight = flags & (Gtk.StateFlags.PRELIGHT)
            is_activated = flags & (Gtk.StateFlags.SELECTED | Gtk.StateFlags.CHECKED)
            is_focused = flags & Gtk.StateFlags.FOCUSED & Gtk.StateFlags.FOCUS_VISIBLE

            revealer.set_reveal_child(not (is_prelight or is_activated or is_focused))

        revealer.set_reveal_child(True)
        widget.connect("state-flags-changed", state_flags_changed, revealer)
        widget.bind_property(
            "visible", revealer, "visible", GObject.BindingFlags.SYNC_CREATE
        )
        widget.bind_property(
            "opacity", revealer, "opacity", GObject.BindingFlags.SYNC_CREATE
        )

    @abstractmethod
    def do_check_value(self) -> None:
        """Check if value exists."""

    @abstractmethod
    def do_update(self) -> None:
        """Update the widget according to the setting."""

    @abstractmethod
    def do_setup(self) -> None:
        """
        Set up the object.

        This is mainly useful to, for example, connect signals, bind properties,
        and other necessary steps when the key is valid.
        """

    def _do_check_key(self) -> None:
        if not self.schema.has_key(self.key):
            logger.warning(f"Key “{self.key}” does not exist. Skipping…")
            raise InvalidSettingError

    def _do_find_schema_id(self) -> None:
        toplevels = Gtk.Window.get_toplevels()

        window = toplevels.get_item(0)

        application = window.get_application()  # type: ignore[union-attr]

        if settings := application.settings.get(self.schema_id):
            self.settings = settings["settings"]
            self.schema = settings["schema"]
            return

        try:
            for source in application.sources:
                self.schema = source.lookup(self.schema_id, True)
                if self.schema:
                    break

            if not self.schema:
                raise InvalidSettingError
        except (AttributeError, TypeError) as error:
            application.settings[self.schema_id] = {"settings": None, "schema": None}
            logger.warning(f"Schema ID “{self.schema_id}” does not exist. Skipping…")
            raise InvalidSettingError(error) from error

        self.settings = Gio.Settings.new(self.schema_id)
        application.settings[self.schema_id] = {
            "settings": self.settings,
            "schema": self.schema,
        }

    def _post_init(self) -> bool:
        try:
            self._do_find_schema_id()
            self._do_check_key()
            self.do_check_value()
            self.do_update()
            self.do_setup()
            self.is_valid_setting = True
        except InvalidSettingError:
            pass
        finally:
            self.bind_property(
                "is-valid-setting",
                self,
                "sensitive",
                GObject.BindingFlags.SYNC_CREATE,
            )

        return False



# reset_button.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from typing import Any

from gi.repository import Gtk


class ResetButton(Gtk.Button):
    """A `Gtk.Button` for resetting a setting."""

    __gtype_name__ = "ResetButton"

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.set_tooltip_text(_("Reset"))
        self.set_icon_name("edit-undo-symbolic")



# adjustment.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, cast

from gi.repository import GLib, GObject, Gio, Gtk

from refine.logger import logger
from refine.widgets.base import BaseInterface
from refine.widgets.reset_button import ResetButton


class Adjustment(Gtk.Adjustment, BaseInterface):
    """
    Derivative of `Gtk.Adjustment` implementing the properties needed to work with `Gio.Settings`.

    This widget is meant to be used in conjunction with widgets relying on `Gtk.Adjustment`.
    """

    __gtype_name__ = "Adjustment"

    settings = cast("Gio.Settings", GObject.Property(type=Gio.Settings))
    schema_id = cast("str", GObject.Property(type=str))
    schema = cast("Gio.SettingsSchema", GObject.Property(type=Gio.SettingsSchema))
    key = cast("str", GObject.Property(type=str))
    is_valid_setting = cast("bool", GObject.Property(type=bool, default=False))

    revealer = GObject.Property(type=Gtk.Revealer)
    reset_button = GObject.Property(type=ResetButton)
    sensitive = GObject.Property(type=bool, default=True)
    restart_required = GObject.Property(type=bool, default=False)

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        BaseInterface.__init__(self, **kwargs)

    def _reset_button_clicked(self, *_args: Any) -> None:
        logger.debug("Revert button pressed")
        self.settings.reset(self.key)
        self.set_value(self.settings.get_value(self.key).unpack())
        self.do_update()

    def _value_changed(self, _adjustment: Gtk.Adjustment, *_args: Any) -> None:
        value_type = self.schema.get_key(self.key).get_value_type()
        format_string = value_type.dup_string()
        variant = GLib.Variant(format_string, self.get_value())

        self.settings.set_value(self.key, variant)
        self.value_set(int(self.get_value()))
        self.do_update()

    def do_setup(self) -> None:
        type Range = tuple[Any, tuple[int, int]]

        schema_key = self.schema.get_key(self.key)
        variant = cast("Range", schema_key.get_range())
        range = variant[1]

        self.set_lower(range[0])
        self.set_upper(range[1])
        self.set_value(self.settings.get_value(self.key).unpack())
        self.do_update()

        self.reset_button.connect("clicked", self._reset_button_clicked)
        self.connect("value-changed", self._value_changed)
        BaseInterface.set_up_separator_revealer(self.revealer, self.reset_button)

    def do_update(self) -> None:
        default_value = cast("GLib.Variant", self.settings.get_default_value(self.key))
        default_int = default_value.unpack()
        value = int(self.get_value())

        self.reset_button.set_sensitive(value != default_int)

    def value_set(self, value: Any) -> None:
        """Log key-value pair."""
        logger.debug(f"Set key “{self.key}” to “{value}”")



# main.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from collections.abc import Callable
from gettext import gettext as _
from typing import Any

import gi

from refine import APP_ID, MAJOR, MINOR, PROFILE

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Pango", "1.0")
gi.require_version("XdpGtk4", "1.0")

# ruff: noqa: E402
from gi.repository import Adw, Gio

from refine import settings
from refine.typing import SettingsSource
from refine.window import Window


class Application(Adw.Application):
    """The main application singleton class."""

    settings: dict[str, SettingsSource] = {}

    def __init__(self) -> None:
        super().__init__(
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.DEFAULT_FLAGS,
            resource_base_path="/page/tesk/Refine",
        )
        self.create_action("quit", lambda *_: self.quit(), ["<primary>q"])
        self.create_action("about", self.on_about_action)

        self.sources: list[Gio.SettingsSchemaSource] = (
            settings.create_schema_sources_list()
        )

    def do_activate(self) -> None:
        """
        Call when the application is activated.

        We raise the application's main window, creating it if necessary.
        """
        win = self.props.active_window
        if not win:
            win = Window(application=self)

        win.present()

        if PROFILE == "development":
            win.add_css_class("devel")

    def on_about_action(self, *_args: Any) -> None:
        """Callback for the app.about action."""
        about = Adw.AboutDialog.new_from_appdata(
            "/page/tesk/Refine/appdata", f"{MAJOR}.{MINOR}.0"
        )

        # Translators: Replace "translator-credits" with your name/username, and optionally an email or URL.
        about.set_translator_credits(_("translator-credits"))
        about.set_copyright(_("Copyright © 2024 Hari Rana (TheEvilSkeleton)"))

        about.add_other_app(
            "com.usebottles.bottles",
            _("Bottles"),
            _("Run Windows software"),
        )
        about.add_other_app(
            "io.gitlab.theevilskeleton.Upscaler",
            _("Upscaler"),
            _("Upscale and enhance images"),
        )

        about.set_developers(
            (
                f"{about.get_developer_name()} https://tesk.page",
                "Evelyn Serra (itsEve) https://itseve.xyz",
                "Mitchell Winkelman (Heliguy) https://heliguy.neocities.org",
            )
        )
        about.set_artists(("Dominik Baran https://gitlab.gnome.org/toad",))
        about.set_debug_info(settings.get_debug_info())
        about.present(self.props.active_window)

    def create_action(
        self,
        name: str,
        callback: Callable[[Any], None],
        shortcuts: None | list[str] = None,
    ) -> None:
        """Add an application action."""
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)


def main() -> int:
    """The application's entry point."""
    app = Application()
    return app.run(sys.argv)



# window_button_row.py
#
# Copyright 2025 Frederick Schenk (CodedOre)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections.abc import Callable
from gettext import gettext as _
from typing import Any, cast

from gi.repository import Adw, GLib, GObject, Gdk, Gio, Gtk

from refine.logger import logger
from refine.typing import LogicalDirectionType
from refine.widgets.base import BaseInterface


class ButtonNotFoundError(AssertionError): ...


class WindowMenuButton(Gtk.Button):
    __gtype_name__ = "WindowMenuButton"

    in_reorderable_box = GObject.Property(type=bool, default=True)
    popover = GObject.Property(type=Gtk.Popover, default=None)

    def __init__(self, button_id: str, **kwargs: Any) -> None:
        super().__init__(
            tooltip_text=button_id.capitalize(),
            **kwargs,
        )
        self.button_id = button_id
        self.add_css_class(button_id)
        self.add_css_class("window-menu-button")

        self.set_child(
            Gtk.Image.new_from_icon_name(f"window-{self.button_id}-symbolic")
        )

        self.popover = Gtk.PopoverMenu()
        self.popover.set_parent(self)

        drag_controller = Gtk.DragSource()
        drag_controller.connect("prepare", self._drag_prepare)
        drag_controller.connect("drag-begin", self._drag_begin)
        drag_controller.connect("drag-cancel", self._drag_end)
        drag_controller.connect("drag-end", self._drag_end)
        self.add_controller(drag_controller)

        self.connect("clicked", self._clicked)

    def _clicked(self, *_args: Any) -> None:
        if self.in_reorderable_box:
            menu = self._reorder_menu()
        else:
            menu = self._hidden_item_menu()

        self.popover.set_menu_model(menu)
        self.popover.popup()

    def _translate_alignment(
        self, alignment: Gtk.DirectionType
    ) -> LogicalDirectionType:
        # fmt: off
        match (alignment, self.get_direction()):
            case (Gtk.DirectionType.LEFT, (Gtk.TextDirection.NONE | Gtk.TextDirection.LTR)):
                return LogicalDirectionType.START
            case (Gtk.DirectionType.LEFT, Gtk.TextDirection.RTL):
                return LogicalDirectionType.END
            case (Gtk.DirectionType.RIGHT, (Gtk.TextDirection.NONE | Gtk.TextDirection.LTR)):
                return LogicalDirectionType.END
            case (Gtk.DirectionType.RIGHT, Gtk.TextDirection.RTL):
                return LogicalDirectionType.START
            case _:
                raise AssertionError
        # fmt: on

    def _hidden_item_menu(self) -> Gio.Menu:
        menu = Gio.Menu()

        menu.append(
            _("Move to _Left Corner"),
            f'window_button_row.show_item(("{self.button_id}",{int(self._translate_alignment(Gtk.DirectionType.LEFT))}))',
        )
        menu.append(
            _("Move to _Right Corner"),
            f'window_button_row.show_item(("{self.button_id}",{int(self._translate_alignment(Gtk.DirectionType.RIGHT))}))',
        )

        return menu

    def _reorder_menu(self) -> Gio.Menu:
        menu = Gio.Menu()

        menu.append(
            _("Move _Left"),
            f'window_button_row.move_item(("{self.button_id}",{int(self._translate_alignment(Gtk.DirectionType.LEFT))}))',
        )
        menu.append(
            _("Move _Right"),
            f'window_button_row.move_item(("{self.button_id}",{int(self._translate_alignment(Gtk.DirectionType.RIGHT))}))',
        )
        menu.append(_("_Hide"), f"window_button_row.hide_item::{self.button_id}")

        return menu

    def get_sibling_button(self, alignment: LogicalDirectionType) -> Gtk.Widget | None:
        """
        Get the next or previous sibling `WindowMenuButton`.

        Depends on the `LogicalDirectionType` provided.
        """
        sibling = self

        while True:
            match alignment:
                case LogicalDirectionType.START:
                    sibling = sibling.get_prev_sibling()  # type: ignore [assignment]
                case LogicalDirectionType.END:
                    sibling = sibling.get_next_sibling()  # type: ignore [assignment]
                case _:
                    raise AssertionError
            if sibling is None:
                return None
            if type(sibling) is WindowMenuButton:
                return sibling

    def _drag_prepare(
        self, _ctrl: Gtk.DragSource, _x: float, _y: float
    ) -> Gdk.ContentProvider:
        return Gdk.ContentProvider.new_for_value(self)

    def _drag_begin(self, ctrl: Gtk.DragSource, _drag: Gdk.Drag) -> None:
        icon = Gtk.WidgetPaintable.new(self)
        ctrl.set_icon(icon, 0, 0)
        self.add_css_class("dimmed")

    def _drag_end(self, *_args: Any) -> None:
        self.remove_css_class("dimmed")


class WindowButtonBin(Adw.Bin):
    __gtype_name__ = "WindowButtonBin"

    button_layout = GObject.Property(type=str, default="")
    available_buttons = GObject.Property(type=GObject.Strv)

    alignment = GObject.Property(type=Gtk.Align, default=Gtk.Align.START)
    reorderable = GObject.Property(type=bool, default=True)

    next_box = GObject.Property(type=Gtk.Widget, default=None)
    prev_box = GObject.Property(type=Gtk.Widget, default=None)

    _unknown_buttons: list[str] = []

    @GObject.Signal
    def changed(self) -> None:
        pass

    def __init__(self, **kwargs: Any) -> None:
        self.set_css_name("windowbuttonbox")
        super().__init__(**kwargs)
        self.connect("notify::button-layout", self._button_layout_set_callback)

        self.button_box = Gtk.Box(css_name="windowcontrols", homogeneous=True)
        self.bind_property("alignment", self.button_box, "halign")
        self.set_child(self.button_box)

        self._drop_place = Gtk.Image.new_from_icon_name("list-drag-handle-symbolic")
        self._drop_place.add_css_class("dimmed")
        self.button_box.append(self._drop_place)

        drop_controller = Gtk.DropTarget.new(
            type=WindowMenuButton, actions=Gdk.DragAction.COPY
        )
        drop_controller.connect("drop", self._on_drop)
        drop_controller.connect("motion", self._on_drag_motion)
        drop_controller.connect("leave", self._check_remaining_handles)
        self.add_controller(drop_controller)

    def _for_each_button(self, func: Callable[[WindowMenuButton], None]) -> None:
        next_child = self.button_box.get_first_child()

        while next_child is not None:
            child = next_child
            next_child = next_child.get_next_sibling()
            if type(child) is WindowMenuButton:
                func(child)

    def _button_layout_set_callback(self, *_args: Any) -> None:
        self._for_each_button(lambda button: self.button_box.remove(button))
        self._unknown_buttons = []

        if len(self.button_layout) != 0:
            buttons = self.button_layout.split(",")
            for button_id in buttons:
                if button_id in self.available_buttons:
                    button = WindowMenuButton(button_id)
                    button.in_reorderable_box = self.reorderable
                    self.button_box.append(button)
                else:
                    self._unknown_buttons.append(button_id)

        self._check_remaining_handles()

    def update_layout(self, *_args: Any) -> None:
        layout = []
        self._for_each_button(lambda button: layout.append(button.button_id))

        if self.alignment is Gtk.Align.END:
            layout = self._unknown_buttons + layout
        else:
            layout = layout + self._unknown_buttons

        self.button_layout = ",".join(layout)

    def _check_remaining_handles(self, *_args: Any) -> None:
        button_count = 0
        child = self.button_box.get_first_child()

        while child is not None:
            if type(child) is WindowMenuButton:
                button_count += 1
            child = child.get_next_sibling()
        self._drop_place.set_visible(button_count == 0)

    def _on_drag_motion(
        self, _ctrl: Gtk.DropTarget, x: float, y: float
    ) -> Gdk.DragAction:
        if self.reorderable:
            self._drop_place.set_visible(True)
            child = self._get_place_sibling(x, y)
            self.button_box.reorder_child_after(self._drop_place, child)
        return Gdk.DragAction.COPY

    def _on_drop(
        self, _ctrl: Gtk.DropTarget, button: WindowMenuButton, x: float, y: float
    ) -> None:
        button.unparent()
        button.in_reorderable_box = self.reorderable

        if self.reorderable:
            sibling = self._get_place_sibling(x, y)
            self.button_box.insert_child_after(button, sibling)
        else:
            self.button_box.append(button)

        self.emit("changed")

    def _get_place_sibling(self, x: float, y: float) -> Gtk.Widget | None:
        child = self.pick(x, y, Gtk.PickFlags.DEFAULT)
        child = self._get_box_child(child)  # type: ignore [arg-type]

        if child is None:
            if self.alignment is Gtk.Align.END:
                return None
            return self.button_box.get_last_child()

        child_width = child.get_width()

        if (x % child_width) > child_width / 2:
            return child
        return child.get_prev_sibling()

    def _get_box_child(self, child: Gtk.Widget) -> Gtk.Widget | None:
        if child is self:
            return None

        while True:
            parent = child.get_parent()
            if parent is None:
                return None
            if parent is self.button_box:
                return child
            child = parent

    def pick_button(self, button_id: str) -> Gtk.Widget | None:
        """Get button widget with the given button ID."""
        child = self.button_box.get_first_child()

        while child is not None:
            child = child.get_next_sibling()
            if type(child) is WindowMenuButton and child.button_id == button_id:
                return child

        return None

    def add_button(self, button: WindowMenuButton) -> None:
        """Append or prepend a button."""
        button.unparent()

        if self.alignment is Gtk.Align.END:
            self.button_box.prepend(button)
        else:
            self.button_box.append(button)

        self.emit("changed")

    def reorder_items(self, widget: Gtk.Widget, sibling: Gtk.Widget) -> None:
        """Reorder buttons and emit the `::changed` signal."""
        self.button_box.reorder_child_after(widget, sibling)
        self.emit("changed")


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/window-buttons-row.ui")
class WindowButtonsRow(Adw.PreferencesRow, BaseInterface):
    __gtype_name__ = "WindowButtonsRow"

    start_box = Gtk.Template.Child()
    end_box = Gtk.Template.Child()
    hidden_box = Gtk.Template.Child()
    reset_button = Gtk.Template.Child()

    settings = cast("Gio.Settings", GObject.Property(type=Gio.Settings))
    schema_id = cast("str", GObject.Property(type=str))
    schema = cast("Gio.SettingsSchema", GObject.Property(type=Gio.SettingsSchema))
    key = cast("str", GObject.Property(type=str))
    is_valid_setting = cast("bool", GObject.Property(type=bool, default=False))

    is_default = GObject.Property(type=bool, default=True)
    button_layout = GObject.Property(type=str, default="")
    title = GObject.Property(type=str, default=None)
    subtitle = GObject.Property(type=str, default=None)
    available_buttons = GObject.Property(type=GObject.Strv)

    def _move_item(self, _widget: Any, _action_name: str, variant: Any) -> None:
        (button_id, alignment) = variant.unpack()
        alignment = LogicalDirectionType(alignment)
        box = None
        button = None

        for searched_box in [self.start_box, self.end_box]:
            button = searched_box.pick_button(button_id)
            if button is not None:
                box = searched_box
                break

        if not box or not button:
            raise IndexError

        button_sibling = button.get_sibling_button(alignment)

        match alignment:
            case LogicalDirectionType.START:
                if button_sibling is not None:
                    box.reorder_items(button_sibling, button)
                if button_sibling is None and box.prev_box is not None:
                    box.prev_box.add_button(button)
            case LogicalDirectionType.END:
                if button_sibling is not None:
                    box.reorder_items(button, button_sibling)
                if button_sibling is None and box.next_box is not None:
                    box.next_box.add_button(button)
            case _:
                logger.error("Unsupported alignment used in move_item")

    def _show_item(self, _widget: Any, _action_name: str, variant: Any) -> None:
        (button_id, alignment) = variant.unpack()
        alignment = LogicalDirectionType(alignment)
        button = self.hidden_box.pick_button(button_id)

        if not button:
            raise ButtonNotFoundError(button_id)

        match alignment:
            case LogicalDirectionType.START:
                self.start_box.add_button(button)
            case LogicalDirectionType.END:
                self.end_box.add_button(button)
            case _:
                logger.error("Unsupported alignment used in show_item")

    def _hide_item(self, _widget: Any, _action_name: str, variant: Any) -> None:
        button_id = variant.unpack()
        button = self.start_box.pick_button(button_id)

        if button is None:
            button = self.end_box.pick_button(button_id)

            if button is None:
                raise ButtonNotFoundError(button_id)

        self.hidden_box.add_button(button)

    def _layout_changed(self, *_args: Any) -> None:
        self.start_box.update_layout()
        self.end_box.update_layout()

        layout = f"{self.start_box.button_layout}:{self.end_box.button_layout}"
        logger.debug(f"Set key “{self.key}” to “{layout}”")

        self.settings.set_value(self.key, GLib.Variant.new_string(layout))

    def _button_layout_set_callback(self, *_args: Any) -> None:
        layout = self.button_layout.split(":")
        self.start_box.button_layout = layout[0]
        self.end_box.button_layout = layout[1]

        hidden_buttons = filter(
            lambda handle: handle not in self.button_layout, self.available_buttons
        )
        self.hidden_box.button_layout = ",".join(hidden_buttons)

        default_value = self.settings.get_default_value(self.key)
        self.is_default = default_value == self.settings.get_value(self.key)

        self.reset_button.set_visible(not self.is_default)

    def _reset_button_clicked(self, *_args: Any) -> None:
        logger.debug("Reset button pressed")
        self.settings.reset(self.key)

    def do_setup(self) -> None:
        self.install_action("window_button_row.show_item", "(si)", self._show_item)
        self.install_action("window_button_row.move_item", "(si)", self._move_item)
        self.install_action("window_button_row.hide_item", "s", self._hide_item)

        self.start_box.connect("changed", self._layout_changed)
        self.end_box.connect("changed", self._layout_changed)
        self.hidden_box.connect("changed", self._layout_changed)

        self.connect("notify::button-layout", self._button_layout_set_callback)
        self.settings.bind(
            self.key, self, "button-layout", Gio.SettingsBindFlags.DEFAULT
        )
        self.reset_button.connect("clicked", self._reset_button_clicked)



# window.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from typing import Any, cast

from gi.repository import Adw, GLib, GObject, Gtk

from refine.logger import logger

# ruff: noqa: F401
from refine.widgets.adjustment import Adjustment
from refine.widgets.base import BaseInterface
from refine.widgets.boolean_switch_row import BooleanSwitchRow
from refine.widgets.combo_row import ComboRow
from refine.widgets.data_combo_row import DataComboRow
from refine.widgets.flags_switch_row import FlagsSwitchRow
from refine.widgets.font_row import FontRow
from refine.widgets.panel_row import PanelRow
from refine.widgets.reset_button import ResetButton
from refine.widgets.value_combo_row import ValueComboRow
from refine.widgets.value_switch_row import ValueSwitchRow
from refine.widgets.window_buttons_row import WindowButtonsRow


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/panel-page.ui")
class PanelPage(Adw.PreferencesPage):
    """Panel page containing options."""

    __gtype_name__ = "PanelPage"


@Gtk.Template.from_resource("/page/tesk/Refine/window.ui")
class Window(Adw.ApplicationWindow):
    """The main window singleton class."""

    __gtype_name__ = "Window"

    stack = Gtk.Template.Child()
    continue_button = Gtk.Template.Child()
    panels_list_box = Gtk.Template.Child()
    unsupported_status_page = Gtk.Template.Child()
    main_navigation_split_view = Gtk.Template.Child()

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self._verify_environment()

        self.needs_restart: list[Gtk.Widget] = []

        panels_schema = (
            "/page/tesk/Refine/sections/appearance.ui",
            "/page/tesk/Refine/sections/mouse-touchpad.ui",
            "/page/tesk/Refine/sections/power.ui",
            "/page/tesk/Refine/sections/shell-compositor.ui",
            "/page/tesk/Refine/sections/sound.ui",
            "/page/tesk/Refine/sections/wellbeing.ui",
        )

        panels_list = Gtk.StringList.new(panels_schema)

        self.panels_list_box.bind_model(panels_list, self._create_row_func)
        self.panels_list_box.connect("row-activated", self._set_preferences_page_cb)

        first_row = self.panels_list_box.get_first_child()
        self._set_preferences_page(first_row)

    def _create_row_func(self, panel: Gtk.StringObject, *_args: Any) -> Gtk.Widget:
        builder = Gtk.Builder.new_from_resource(panel.get_string())
        row = PanelRow()
        page = builder.get_object("page")

        row.page = page  # type: ignore [method-assign]

        row.page.bind_property(
            "icon-name", row, "icon-name", GObject.BindingFlags.SYNC_CREATE
        )
        row.page.bind_property("title", row, "title", GObject.BindingFlags.SYNC_CREATE)

        return row

    def _set_preferences_page(self, row: PanelRow, *_args: Any) -> None:
        navigation_page = self.main_navigation_split_view.get_content()
        toolbar_view = navigation_page.get_child()

        toolbar_view.set_content(row.page)
        navigation_page.set_title(row.page.get_title())

    def _set_preferences_page_cb(self, _: Any, row: PanelRow, *_args: Any) -> None:
        self._set_preferences_page(row)
        self.main_navigation_split_view.set_show_content(True)

    def _verify_environment(self) -> None:
        if GLib.getenv("XDG_CURRENT_DESKTOP") != "GNOME":
            self.stack.set_visible_child_name("unsupported-environment-page")
            self.continue_button.connect(
                "clicked", lambda _: self.stack.set_visible_child_name("refine-page")
            )

            unintended_behavior_label = _(
                "Some settings may not work as intended and could lead to undesirable behavior."
            )
            unintended_behavior_label = f"<b>{unintended_behavior_label}</b>"

            if GLib.getenv("XDG_SESSION_DESKTOP") == "gnome":
                description = (
                    _(
                        "Refine only supports versions of GNOME that have not been modified by the operating system."
                    )
                    + " "
                    + unintended_behavior_label
                )
            else:
                description = (
                    _(
                        "Refine is only intended to run on the GNOME desktop environment."
                    )
                    + " "
                    + unintended_behavior_label
                )

            self.unsupported_status_page.set_description(description)

    # TODO: Find a way to request a logout dialog
    @GObject.Signal
    def restart_items_changed(self, widget: Gtk.Widget) -> None:
        logger.info(f"{self.__class__.__name__}::restart-items-changed signal emitted")

        if not (panel := cast("PanelPage", widget.get_ancestor(PanelPage))):
            return

        try:
            if not (banner := panel.get_banner()):
                return
        except (AttributeError, TypeError):
            return

        try:
            self.needs_restart.remove(widget)
            banner.set_revealed(False)
        except ValueError:
            self.needs_restart.append(widget)
            banner.set_revealed(True)



# font_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import TYPE_CHECKING, Any, cast

from gi.repository import GLib, GObject, Gio, Gtk, Pango

from refine.logger import logger
from refine.widgets.action_row import ActionRow
from refine.widgets.base import BaseInterface

if TYPE_CHECKING:
    from gi.repository import Adw


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/font-row.ui")
class FontRow(ActionRow):
    """
    An `ActionRow` for displaying and interacting with font.

    It is also used for launching the `Gtk.FontDialog`.
    """

    __gtype_name__ = "FontRow"

    revert_button = Gtk.Template.Child()
    revealer = Gtk.Template.Child()

    is_default = GObject.Property(type=bool, default=True)

    def _row_activated(self, *_args: Any) -> None:
        def choose_font_and_features_callback(
            _dialog: Gtk.FontDialog, result: Gio.AsyncResult
        ) -> None:
            try:
                font = font_dialog.choose_font_and_features_finish(result)
                font_string = font[1].to_string()
            except (IndexError, GLib.Error):
                return

            self.settings.set_string(self.key, font_string)
            self.value_set(self.settings.get_string(self.key))
            self.do_update()

        font_string = self.settings.get_string(self.key)
        font_description = Pango.FontDescription.from_string(font_string)
        font_dialog = Gtk.FontDialog()

        if window := cast("Adw.Window | None", self.get_root()):
            font_dialog.choose_font_and_features(
                window, font_description, callback=choose_font_and_features_callback
            )

    def _revert_button_clicked(self, *_args: Any) -> None:
        logger.debug("Revert button pressed")
        self.settings.reset(self.key)
        self.do_update()

    def do_update(self) -> None:
        font_string = self.settings.get_string(self.key)
        default_value = cast("GLib.Variant", self.settings.get_default_value(self.key))
        default_font_string = default_value.get_string()

        self.is_default = default_font_string != font_string

        font_description = Pango.FontDescription.from_string(font_string)
        font_family = font_description.get_family()

        self.set_subtitle(f"<span face='{font_family}'>{font_string}</span>")

    def do_setup(self) -> None:
        self.connect("activated", self._row_activated)
        self.bind_property(
            "is-default",
            self.revert_button,
            "visible",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self.revert_button.connect("clicked", self._revert_button_clicked)
        BaseInterface.set_up_separator_revealer(self.revealer, self.revert_button)



# boolean_switch_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any

from gi.repository import Gio, Gtk

from refine.logger import logger
from refine.widgets.base import BaseInterface
from refine.widgets.toggle_row import ToggleRow


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/boolean-switch-row.ui")
class BooleanSwitchRow(ToggleRow):
    """A `ToggleRow` with on/off switches."""

    __gtype_name__ = "BooleanSwitchRow"

    revert_button = Gtk.Template.Child()
    revealer = Gtk.Template.Child()

    def do_setup(self) -> None:
        super().do_setup()
        self.settings.bind(self.key, self, "active", Gio.SettingsBindFlags.DEFAULT)

        self.connect("notify::active", self._active_set_callback)
        self.revert_button.connect("clicked", self._revert_button_clicked)
        BaseInterface.set_up_separator_revealer(self.revealer, self.revert_button)

    def do_update(self) -> None:
        super().do_update()
        self.is_default = self.settings.get_default_value(
            self.key
        ) != self.settings.get_value(self.key)

        self.revert_button.set_sensitive(self.is_default)

    def _active_set_callback(self, *_args: Any) -> None:
        self.do_update()
        self.value_set(self.settings.get_value(self.key))

    def _revert_button_clicked(self, *_args: Any) -> None:
        logger.debug("Revert button pressed")
        self.settings.reset(self.key)



# toggle_row.py
#
# Copyright 2025 Hari Rana / TheEvilSkeleton <theevilskeleton@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, Gtk

from refine.widgets.action_row import ActionRow


class ToggleRow(ActionRow):
    __gtype_name__ = "ToggleRow"

    active = GObject.Property(type=bool, default=False)
    is_default = GObject.Property(type=bool, default=True)

    def _update_accessible_state(self) -> None:
        self.update_state(
            *zip(  # type: ignore [arg-type]
                (Gtk.AccessibleState.CHECKED, int(self.active)),
            )
        )

    def do_setup(self) -> None:
        self.props.accessible_role = Gtk.AccessibleRole.SWITCH
        self._update_accessible_state()

    def do_update(self) -> None:
        self._update_accessible_state()



# settings.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from pathlib import Path

from gi.repository import GLib, Gio, Xdp


def get_user_config_dir() -> str:
    """Retrieve the user's config directory."""
    if not Xdp.Portal.running_under_sandbox():
        return GLib.get_user_config_dir()

    return os.environ.get(
        "HOST_XDG_CONFIG_HOME", str(Path(GLib.get_home_dir(), "." + "config"))
    )


def get_user_data_dir() -> str:
    """Retrieve the user's data directory."""
    if not Xdp.Portal.running_under_sandbox():
        return GLib.get_user_data_dir()

    return os.environ.get(
        "HOST_XDG_DATA_HOME",
        str(Path(GLib.get_home_dir(), "." + "local", "share")),
    )


def get_system_data_dirs() -> list[str]:
    """Retrieve the system's data directory."""
    data_dirs = GLib.get_system_data_dirs()

    if Xdp.Portal.running_under_flatpak():
        data_dirs.remove(str(Path(os.sep, "app", "share")))

    return data_dirs


def get_debug_info() -> str:
    """Retrieve important information for debugging."""
    environment_variables = (
        "XDG_DATA_DIRS",
        "XDG_CONFIG_HOME",
        "XDG_DATA_HOME",
        "HOST_XDG_CONFIG_HOME",
        "HOST_XDG_DATA_HOME",
    )

    home = Path(GLib.get_home_dir()).name
    output = ""
    for environment_variable in environment_variables:
        original = f"{environment_variable}: “{GLib.getenv(environment_variable)}”\n"
        output += original.replace(home, "[REDACTED]")

    return str(output)


def create_schema_sources_list() -> list[Gio.SettingsSchemaSource]:
    """Create a list schema sources based on the user's system data directory."""
    sources = []

    for data_dir in GLib.get_system_data_dirs():
        path = Path(data_dir, "glib-2.0", "schemas")
        if path.is_dir():
            source = Gio.SettingsSchemaSource.new_from_directory(str(path), None, True)
            sources.append(source)

    return sources



# panel_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


from gi.repository import Adw, GObject, Gtk


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/panel-row.ui")
class PanelRow(Adw.PreferencesRow):
    """An `Adw.PreferencesRow` for displaying information of a page."""

    __gtype_name__ = "PanelRow"

    @GObject.Property(type=Adw.PreferencesPage)
    def page(self) -> Adw.PreferencesPage:
        """Get or set the `Adw.PreferencesPage` of the row."""
        return self._page

    @page.setter  # type: ignore [no-redef]
    def page(self, page: Adw.PreferencesPage) -> None:
        self._page = page

    @GObject.Property(type=str, default="")
    def icon_name(self) -> str:
        """Get or set the icon name."""
        return self._icon_name

    @icon_name.setter  # type: ignore [no-redef]
    def icon_name(self, icon_name: str) -> None:
        self._icon_name = icon_name



# combo_row.py
#
# Copyright 2025 Mitchell Winkelman (Heliguy)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import TYPE_CHECKING, Any, cast

from gi.repository import Adw, GObject, Gio, Gtk

from refine.logger import logger
from refine.widgets.base import BaseInterface

if TYPE_CHECKING:
    from gi.repository import GLib


class ComboRow(Adw.ComboRow, BaseInterface):
    """
    Base combo row widget implementing the `schema-id` and `key` properties.

    It is assumed to derive from it, rather than using it directly.
    """

    __gtype_name__ = "ComboRow"

    is_valid_setting = cast("bool", GObject.Property(type=bool, default=False))
    settings = cast("Gio.Settings", GObject.Property(type=Gio.Settings))
    schema = cast("Gio.SettingsSchema", GObject.Property(type=Gio.SettingsSchema))
    schema_id = cast("str", GObject.Property(type=str))
    key = cast("str", GObject.Property(type=str))

    restart_required = GObject.Property(type=bool, default=False)
    is_default = GObject.Property(type=bool, default=True)

    original_list: list[str]
    transformed_list: list[str]
    revert_button: Gtk.Button
    revealer: Gtk.Revealer

    def _settings_changed(self, *_args: Any) -> None:
        row_string = self.original_list[self.get_selected()]
        settings_string = self.settings.get_string(self.key)

        if row_string != settings_string:
            self.emit("inconsistent")

    def _inconsistent_set(self, *_args: Any) -> None:
        self.handler_block_by_func(self.selected_callback)  # type: ignore [no-untyped-call]

        string = self.settings.get_string(self.key)

        self.set_selected(self.original_list.index(string))
        self.set_is_default(self.key)

        self.handler_unblock_by_func(self.selected_callback)  # type: ignore [no-untyped-call]

    def _reset_button_clicked(self, *_args: Any) -> None:
        logger.debug("Revert button pressed")

        string_object = cast(
            "Gtk.StringObject", self.settings.get_default_value(self.key)
        )
        if not string_object:
            return

        string = string_object.get_string()

        self.settings.reset(self.key)
        self.set_selected(self.original_list.index(string))
        self.set_is_default(self.key)

    def set_is_default(self, key: str) -> None:
        """Update the `ComboRow.is-default` property."""
        default_value = cast("GLib.Variant", self.settings.get_default_value(key))
        default_string = default_value.get_string()

        row_string = self.original_list[self.get_selected()]

        self.is_default = row_string == default_string

    def selected_callback(self, *_args: Any) -> None:
        """Callback method for selecting an item."""
        row_string = self.original_list[self.get_selected()]

        self.set_is_default(self.key)
        self.settings.set_string(self.key, row_string)
        self.value_set(row_string)

    def value_set(self, value: Any) -> None:
        """Log key-value pair."""
        logger.debug(f"Set key “{self.key}” to “{value}”")

        if not self.restart_required:
            return

        toplevels = Gtk.Window.get_toplevels()

        if not (window := toplevels.get_item(0)):
            return

        window.emit("restart-items-changed", self)

    def do_setup(self) -> None:
        self.revert_button.connect("clicked", self._reset_button_clicked)
        self.bind_property(
            "is-default",
            self.revert_button,
            "sensitive",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
        )
        BaseInterface.set_up_separator_revealer(self.revealer, self.revert_button)
        self.connect("inconsistent", self._inconsistent_set)
        self.settings.connect("changed", self._settings_changed)

    @GObject.Signal
    def inconsistent(self) -> None:
        """Log inconsistency signal."""
        logger.info(f"{self.__class__.__name__}::inconsistent signal emitted")



# value_switch_row.py
#
# Copyright 2024 Evelyn Serra (itsEve)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any

from gi.repository import Gtk

from refine.widgets.string_row import StringRow


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/value-switch-row.ui")
class ValueSwitchRow(StringRow):
    """A `StringRow` to enable or disable a feature from a combo action."""

    __gtype_name__ = "ValueSwitchRow"

    def get_string(self) -> str:
        """Retrieve string from key."""
        return self.settings.get_string(self.key)

    def active_set(self, *_args: Any) -> None:
        """Set whether @self is in its “on” or “off” position."""
        self.do_update()
        if self.active:
            self.settings.set_string(self.key, self.value)
        else:
            self.settings.reset(self.key)

        self.value_set(self.settings.get_string(self.key))



# string_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, cast

from gi.repository import GObject

from refine.logger import logger
from refine.widgets.toggle_row import ToggleRow


class StringRow(ToggleRow):
    """A `ToggleRow` adding the base properties and functionality to manipulate strings and string lists."""

    __gtype_name__ = "StringRow"

    value = GObject.Property(type=str)

    def _settings_changed(self, *_args: Any) -> None:
        flag = self.get_string()
        is_setting_active = cast("str", self.value) in flag

        if self.active != is_setting_active:
            self.emit("inconsistent")

    def _inconsistent_set(self, *_args: Any) -> None:
        self.handler_block_by_func(self.active_set)  # type: ignore [no-untyped-call]
        flag = self.get_string()
        self.active = cast("str", self.value) in flag
        self.value_set(flag)
        self.handler_unblock_by_func(self.active_set)  # type: ignore [no-untyped-call]

    def get_string(self) -> str | list[str]:
        """String or list getter virtual method."""
        raise NotImplementedError

    def active_set(self, *_args: Any) -> None:
        """Active setter virtual method."""
        raise NotImplementedError

    def do_setup(self) -> None:
        super().do_setup()
        self.active = cast("str", self.value) in self.get_string()
        self.connect("notify::active", self.active_set)
        self.connect("inconsistent", self._inconsistent_set)
        self.settings.connect("changed", self._settings_changed)

    @GObject.Signal
    def inconsistent(self) -> None:
        """Log inconsistency signal."""
        logger.info(f"{self.__class__.__name__}::inconsistent signal emitted")



# data_combo_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from pathlib import Path

from gi.repository import GObject, Gtk

from refine import settings
from refine.widgets.base import InvalidSettingError
from refine.widgets.combo_row import ComboRow


@Gtk.Template.from_resource("/page/tesk/Refine/widgets/data-combo-row.ui")
class DataComboRow(ComboRow):
    """An `Adw.ComboRow` implementing the properties needed to work with `Gio.Settings`."""

    __gtype_name__ = "DataComboRow"

    revert_button = Gtk.Template.Child()
    revealer = Gtk.Template.Child()

    require_file = GObject.Property(type=str, default="")
    lookup_directory = GObject.Property(type=str, default="")
    data_subdirectory = GObject.Property(type=str, default="")

    aliases_dict: dict[str, str] = {}

    def do_setup(self) -> None:
        if not (string := self.settings.get_string(self.key)):
            raise InvalidSettingError

        system_data_dirs_list: list[Path] = []
        for directory in settings.get_system_data_dirs():
            found_directories = Path(directory, self.data_subdirectory).glob("*")
            system_data_dirs_list += found_directories

        data_dirs_list = system_data_dirs_list

        data_dirs_list += Path(
            settings.get_user_data_dir(), self.data_subdirectory
        ).glob("*")

        data_dirs_list = [
            path
            for path in data_dirs_list
            if Path(path, self.lookup_directory).is_dir()
        ]

        data_dirs_list = [
            path
            for path in data_dirs_list
            if Path(path, self.lookup_directory, self.require_file).exists()
        ]

        data_dirs_list = [path.name for path in data_dirs_list if path.is_dir()]

        if self.aliases_dict:
            data_dirs_list = [
                self.aliases_dict.get(str(path), path) for path in data_dirs_list
            ]

        self.original_list = self.transformed_list = sorted(
            {str(path) for path in data_dirs_list}, key=str.lower
        )

        data_dirs_model = Gtk.StringList.new(self.transformed_list)
        self.set_model(data_dirs_model)

        self.set_selected(self.original_list.index(string))
        self.set_is_default(self.key)
        self.connect("notify::selected", self.selected_callback)

        super().do_setup()

    @GObject.Property(type=str)
    def aliases(self) -> dict[str, str]:
        """Get a dictionary of aliases or set a newline-separated string that gets converted to a dictionary."""
        return self.aliases_dict

    @aliases.setter  # type: ignore [no-redef]
    def aliases(self, aliases: str) -> None:
        aliases_list = aliases.split()

        # Credit: https://stackoverflow.com/a/12739974
        self.aliases_dict = dict(
            zip(aliases_list[::2], aliases_list[1::2], strict=False)
        )



# action_row.py
#
# Copyright 2024 Hari Rana (TheEvilSkeleton)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, cast

from gi.repository import Adw, GObject, Gio, Gtk

from refine.logger import logger
from refine.widgets.base import BaseInterface


class ActionRow(Adw.ActionRow, BaseInterface):
    """
    Base action row widget implementing the `schema-id` and `key` properties.

    It is assumed to derive from it, rather than using it directly.
    """

    __gtype_name__ = "ActionRow"

    schema = cast("Gio.SettingsSchema", GObject.Property(type=Gio.SettingsSchema))
    schema_id = cast("str", GObject.Property(type=str))
    key = cast("str", GObject.Property(type=str))
    is_valid_setting = cast("bool", GObject.Property(type=bool, default=False))

    restart_required = GObject.Property(type=bool, default=False)

    def value_set(self, value: Any) -> None:
        """Log key-value pair."""
        logger.debug(f"Set key “{self.key}” to “{value}”")

        if not self.restart_required:
            return

        toplevels = Gtk.Window.get_toplevels()

        if not (window := toplevels.get_item(0)):
            return

        window.emit("restart-items-changed", self)
